package de.unikoblenz.fgbks.core.dmn.verification;

import de.unikoblenz.fgbks.core.dmn.utils.DmnService;
import de.unikoblenz.fgbks.core.dmn.verification.result.VerifierResult;
import de.unikoblenz.fgbks.core.dmn.verification.result.VerifierResultSet;
import de.unikoblenz.fgbks.core.dmn.verification.result.VerifierResultSet.Builder;
import de.unikoblenz.fgbks.core.dmn.verification.verifier.AbstractVerifier;
import de.unikoblenz.fgbks.core.dmn.verification.verifier.DmnVerifier;
import de.unikoblenz.fgbks.core.dmn.verification.verifier.classification.Classification;
import de.unikoblenz.fgbks.core.dmn.verification.verifier.classification.ClassificationType;
import de.unikoblenz.fgbks.core.dmn.verification.verifier.config.AbstractDmnVerifierConfig;
import de.unikoblenz.fgbks.core.dmn.verification.verifier.config.DefaultConfiguration;
import de.unikoblenz.fgbks.core.dmn.verification.verifier.types.Type;
import de.unikoblenz.fgbks.core.dmn.verification.verifier.types.VerificationType;
import java.lang.reflect.InvocationTargetException;
import java.util.ArrayList;
import java.util.Collections;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.ExecutionException;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import java.util.concurrent.Future;
import javax.enterprise.context.ApplicationScoped;
import javax.inject.Inject;
import org.apache.commons.lang3.Validate;
import org.eclipse.microprofile.config.inject.ConfigProperty;
import org.reflections.Reflections;

@ApplicationScoped
public class DmnVerificationService {

  @ConfigProperty(name = "verifier.threads", defaultValue = "2")
  Integer maxThreads;

  @ConfigProperty(
      name = "verifier.package",
      defaultValue = "de.unikoblenz.fgbks.core.dmn.verification.verifier")
  String packagePath;

  private ExecutorService executor;
  private Set<Class<?>> verifierClasses;
  private List<VerificationType> verificationTypes;
  private List<ClassificationType> verificationClassificationTypes;

  @Inject protected DmnService dmnService;

  protected DmnVerificationService() {
    super();
  }

  private void initVerifier() {
    if (executor == null || verifierClasses == null) {
      System.out.println("Creating thread pool with " + maxThreads + " threads!"); // TODO remove
      executor = Executors.newFixedThreadPool(maxThreads);
      Reflections reflections = new Reflections(packagePath);
      verifierClasses = reflections.getTypesAnnotatedWith(DmnVerifier.class);

      verificationTypes = new ArrayList<>();
      Set<Class<?>> clVts = reflections.getTypesAnnotatedWith(Type.class);
      for (Class<?> clVt : clVts) {
        try {
          verificationTypes.add(
              (VerificationType) clVt.getMethod("getInstance").invoke(null, null));
        } catch (IllegalAccessException | NoSuchMethodException | InvocationTargetException e) {
          e.printStackTrace();
        }
      }

      verificationClassificationTypes = new ArrayList<>();
      Set<Class<?>> clCts = reflections.getTypesAnnotatedWith(Classification.class);
      for (Class<?> clCt : clCts) {
        try {
          verificationClassificationTypes.add(
              (ClassificationType) clCt.getMethod("getInstance").invoke(null, null));
        } catch (IllegalAccessException | NoSuchMethodException | InvocationTargetException e) {
          e.printStackTrace();
        }
      }
    }
  }

  public VerifierResultSet generate(String dmnXml) {
    return generateWithFilter(dmnXml, null, null);
  }

  public VerifierResultSet generateFromType(String type, String dmnXml) {
    return generateWithFilter(dmnXml, type, null);
  }

  public VerifierResultSet generateFromClassification(String classification, String dmnXml) {
    return generateWithFilter(dmnXml, null, classification);
  }

  private VerifierResultSet generateWithFilter(
      String dmnXml, String typeName, String classificationName) {
    initVerifier();
    Validate.notNull(dmnXml);
    Map<Class<? extends AbstractDmnVerifierConfig>, AbstractDmnVerifierConfig> configs =
        getConfigs(dmnXml);
    Builder resultBuilder = VerifierResultSet.getBuilder();

    List<Future<VerifierResult>> results = new ArrayList<>(verifierClasses.size());

    for (Class<?> verifierClass : verifierClasses) {
      AbstractVerifier av = null;
      try {
        av = (AbstractVerifier) verifierClass.newInstance();
      } catch (InstantiationException | IllegalAccessException e) {
        e.printStackTrace();
      }
      if (av != null
          && (typeName == null || av.getVerificationType().getName().getValue().equals(typeName))
          && (classificationName == null
              || av.getVerificationType()
                  .getClassification()
                  .getName()
                  .getValue()
                  .equals(classificationName))) {
        // Create Configuration
        AbstractDmnVerifierConfig vConfig =
            configs.get(av.getClass().getAnnotation(DmnVerifier.class).verifierConfig());

        // Add Configuration and start Calculation
        results.add(av.withConfiguration(vConfig).verify(executor));
      }
    }

    results.forEach(
        r -> {
          try {
            resultBuilder.addVerifierResult(r.get());
          } catch (InterruptedException e) {
            e.printStackTrace();
          } catch (ExecutionException e) {
            e.printStackTrace();
          }
        });

    return resultBuilder.build();
  }

  private Map<Class<? extends AbstractDmnVerifierConfig>, AbstractDmnVerifierConfig> getConfigs(
      String dmnXML) {
    Map<Class<? extends AbstractDmnVerifierConfig>, AbstractDmnVerifierConfig> map =
        new ConcurrentHashMap<>();
    map.put(
        DefaultConfiguration.class,
        DefaultConfiguration.getBuilder()
            .withDmnObjectContainer(dmnService.getDmnServiceFromXml(dmnXML))
            .build());
    return map;
  }

  public List<VerificationType> getVerificationTypes() {
    initVerifier();
    return Collections.unmodifiableList(verificationTypes);
  }

  public List<ClassificationType> getVerificationClassificationTypes() {
    initVerifier();
    return Collections.unmodifiableList(verificationClassificationTypes);
  }
}
